﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Diagnostics;
using KernCADnet;

namespace SplineMods
{
    public partial class SplineModsForm : Form
    {
        enum State { eStateIdle, eStatePickingPoint, eStatePickingMoveToPoint, eStateReadyToMove, eStateDone };

        State m_state = State.eStateIdle;
        IObjectGenerator_DG m_iGener;	// Generic constructor
        IModel_DG m_iModel;
        IStdShape_DG m_iStdShape;
        IPick m_iPick;
        IEntity_DG m_iEntity;			// Entity which has the curve
        IBSplineCurve_DG m_iSplineCurve;
        DIPoint m_pt;			// The picked point to be modified
        DIPoint m_ptNew;		// New modified position of the picked point
        IPointArray_DG m_iArrPnt;	// Definition points for the spline

        public SplineModsForm()
        {
            InitializeComponent();
        }

        private void OnFormLoad(object sender, EventArgs e)
        {
            m_iModel = m_dgk.GetModel() as IModel_DG;
            m_iGener = m_iModel as IObjectGenerator_DG;
            m_iStdShape = m_iModel as IStdShape_DG;

            CreateSpline();

            // Add polygon of poles to the scene
            DisplayPolesPolygon(m_iSplineCurve);
            DisplayInterpolationPoints(m_iArrPnt);		// Show the points

            ZoomExtents(2.5);
            UpdatePage();
        }

        void ZoomExtents(double zoom)
        {
            IView_DG iView = m_dgk.GetView() as IView_DG;
	        // Fit into the window
            iView.Reset(true, true);
	        // Zoom out
		    iView.Zoom(zoom);
        }

        void CreateSpline()
        {
            if (m_iArrPnt == null)
                m_iArrPnt = m_iGener.Create("PointArray_DG") as IPointArray_DG;

            m_iArrPnt.SetCount(5);
            PointDg pt = new PointDg();
            pt.Set( 0, 0, 10);  m_iArrPnt.SetPoint(0, pt);
            pt.Set(10, 20, 20); m_iArrPnt.SetPoint(1, pt);
            pt.Set(20, 30, 30); m_iArrPnt.SetPoint(2, pt);
            pt.Set(40, 30, 40); m_iArrPnt.SetPoint(3, pt);
            pt.Set(50, 50, 50); m_iArrPnt.SetPoint(4, pt);

            m_iSplineCurve = m_iGener.Create("BSplineCurve_DG") as IBSplineCurve_DG;

            m_iSplineCurve.InitFromPoints(m_iArrPnt, 3, 8, Continuity_DG.eCont_C2, 1e-3);

            IBRepShape_DG iShape = m_iGener.Create1("BRepEdge_DG", m_iSplineCurve) as IBRepShape_DG;

            // Add shape to the displayed model
            IEntity_DG entity = m_iModel.AddBRepShape(iShape);

            entity.SetColor(1, 0, 0);		//Red
        }

        void DisplayPolesPolygon(IBSplineCurve_DG iCurve)
        {
	        IPointArray_DG iArrPnt = iCurve.GetPoles();

	        int n = iArrPnt.GetCount();
            // Create segment strip geometry and entity
	        IEntity_DG iEntityPolygon = m_iModel.AddNewEntity1("SegmentStrip");
	        iEntityPolygon.SetColor(0, 0, 1);	//Blue in rgb

	        // Copy points
	        ISectionPointSet iPointSet = iEntityPolygon as ISectionPointSet;
	        IArray2 iArray = iPointSet as IArray2;
	        iArray.SetCount(n);
	        IMatrixData points = iPointSet.GetPoints();

	        for (int k = 0; k<n; k++)
	        {
		        Point_DG pt = iArrPnt.GetPoint(k);
		        points.Set(k, 0, pt.x[0]);
		        points.Set(k, 1, pt.x[1]);
		        points.Set(k, 2, pt.x[2]);
	        }
        }

        void AddSpherePoint(Point_DG pt, String color)
        {
	        IEntity_DG pEnt = m_iStdShape.Sphere(0.5);
	        IFrame_DG iFrame = pEnt.GetLocation();
	        iFrame.SetOrigin(pt);
	        if(color == "Yellow")
		        pEnt.SetColor(1.0f, 1.0f, 0);
	        else if (color == "Red")
		        pEnt.SetColor(1.0f, 0.0f, 0);
	        else if (color == "Blue")
		        pEnt.SetColor(0.0f, 0.0f, 1.0f);
	        m_iModel.AddEntity(pEnt);
        }

        void DisplayInterpolationPoints(IPointArray_DG iArrPnt)
        {
	        int n = iArrPnt.GetCount();
            // Create point set geometry and entity
	        IEntity_DG iEntityPoints = m_iModel.AddNewEntity1("PointArray");

	        // Copy points
            ISectionPointSet iPointSet = iEntityPoints as ISectionPointSet;
            IArray2 iArray = iPointSet as IArray2;
	        iArray.SetCount(n);
	        IMatrixData points = iPointSet.GetPoints();

	        for (int k = 0; k<n; k++)
	        {
                Point_DG pt = iArrPnt.GetPoint(k);
		        points.Set(k, 0, pt.x[0]);		
		        points.Set(k, 1, pt.x[1]);		
		        points.Set(k, 2, pt.x[2]);

		        AddSpherePoint(pt, "Yellow");
	        }

	        iPointSet.PointSize = 5;		// In pixels
        }

        private void OnDgkMouseDown(object sender, AxKernCADnet._DKernCADnetEvents_MouseDownEvent e)
        {
            if (m_state == State.eStatePickingPoint)
            {
                if (m_iPick == null)
                {
                    IView_DG iView = m_dgk.GetView() as IView_DG;
                    m_iPick = iView as IPick;
                }

                IItem item = m_iPick.GetItemAt3D(e.x, e.y, out m_pt) as IItem;

                AddSpherePoint(new PointDg(m_pt.x[0], m_pt.x[1], m_pt.x[2]), "Red");		// For display

                m_iEntity = item as IEntity_DG;

                m_dgk.UpdateView();

                m_state = State.eStatePickingMoveToPoint;
            }
            else if (m_state == State.eStatePickingMoveToPoint)
            {
                Debug.Assert(m_iPick!=null);

                IItem item = m_iPick.GetItemAt3D(e.x, e.y, out m_ptNew) as IItem;

                PointDg ptNew = new PointDg(m_ptNew.x[0], m_ptNew.x[1], m_ptNew.x[2]);

                AddSpherePoint(ptNew, "Blue");		// For display

                m_dgk.UpdateView();

                m_state = State.eStateReadyToMove;
            }

            UpdatePage();
        }

        void ModifySpline()
        {
	        ICurve_DG iCurve = m_iSplineCurve as ICurve_DG;
	        
	        Point_DG pt = new PointDg(m_pt.x[0], m_pt.x[1], m_pt.x[2]);
            Point_DG ptNearest;
	        double u = iCurve.GetNearestPoint(pt, out ptNearest);

	        int i1 = FindPoleSegment(iCurve, u);					Debug.Assert(i1>=0);

	        PointDg ptNew = new PointDg(m_ptNew.x[0], m_ptNew.x[1], m_ptNew.x[2]);

            int k0, k1;
            m_iSplineCurve.ModifyPoint(u, ptNew, i1 - 1, i1, out k0, out k1);

	        // Set modified flag. Curve is too low-level attribute to know how to notify entity about the modification
	        IGeometry_DG iGeometry = m_iEntity.GetGeometry();
	        iGeometry.SetModified();
        }

        // Find two sequential poles adjacent to the picked point. Return index of the second
        int FindPoleSegment(ICurve_DG iCurve, double u)
        {
	        int n = m_iArrPnt.GetCount();

            Point_DG ptNearest;
	        // Find the first definition point with parameter greater than u
	        // We also could have used projections of poles (nearest point) to the curve

	        for (int i=0; i<n; i++)
	        {
                PointDg pt = m_iArrPnt.GetPoint(i);
                double t = iCurve.GetNearestPoint(pt, out ptNearest);
		        if (t >= u)
			        return i;
	        }
	        return -1;
        }

        private void OnStep0Begin(object sender, EventArgs e)
        {
            if (m_state == State.eStateDone)
            {
                Close();
                return;
            }

            m_state = State.eStatePickingPoint;
            UpdatePage();
        }

        void UpdatePage()
        {
            buttonBegin.Enabled = m_state == State.eStateIdle || m_state == State.eStateDone;
            buttonStep1.Enabled = m_state == State.eStatePickingPoint;
            buttonStep2.Enabled = m_state == State.eStatePickingMoveToPoint;
            buttonStep3Modify.Enabled = m_state == State.eStateReadyToMove;

            if (m_state == State.eStateDone)
                buttonBegin.Text = "Please restart to repeat";
        }

        private void OnStep3Modify(object sender, EventArgs e)
        {
            if (m_iEntity != null)			// Otherwise have not picked point on the first staep
            {
                ModifySpline();
                m_dgk.UpdateView();
            }

            m_state = State.eStateDone;
            UpdatePage();
        }
    }
}
